"""Manipulation of multiple Trello boards at once."""

# import collections
import itertools
import logging
import typing as t

import trello  # py-trello package

from .board_cache import BoardKind, BOARD_SYMBOLS, BOARD_LISTS, BoardCache
from .recurring_board_cache import RecurringBoardCache

_LOG = logging.getLogger(__name__)

_MANIPULATOR_SYMBOL = '⚙'


class TrelloManipulator(trello.TrelloClient):
    """Client for Trello with high-level functionality."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._warmup_board = None
        self._work_board = None
        # self._cache = False
        self._boards = None
        self._boards_by_id = None
        self._boards_normal = []
        self._boards_recurring = []

    @property
    def warmup_board(self):
        return self._warmup_board

    @property
    def work_board(self):
        return self._work_board

    # @property
    # def cache(self) -> bool:
    #     return self._cache

    # @cache.setter
    # def cache(self, value: bool):
    #     self._cache = value

    @property
    def boards(self):
        # if not self._cache or self._boards is None:
        #     self.refresh_boards_list()
        return self._boards

    @property
    def normal_boards(self):
        return self._boards_normal

    @property
    def recurring_boards(self):
        return self._boards_recurring

    def refresh_boards_list(self):
        self._boards = self.list_boards()
        self._boards_by_id = {b.id: b for b in self._boards}

    def refresh_handled_boards_cache(self):
        self._warmup_board.refresh_cache()
        self._work_board.refresh_cache()
        for board in itertools.chain(self._boards_normal, self._boards_recurring):
            board.refresh_cache()

    def refresh_cache(self):
        _LOG.info('completely rebuilding cache...')
        self.refresh_boards_list()
        self.refresh_handled_boards_cache()
        _LOG.info('cache was completely rebuilt.')

    def find_all_board_ids(self, param_name: str, param_value: t.Any) -> list:
        return [_.id for _ in self._boards if getattr(_, param_name) == param_value]

    def find_board_id(self, param_name: str, param_value: t.Any) -> str:
        board_ids = self.find_all_board_ids(param_name, param_value)
        assert len(board_ids) == 1, len(board_ids)
        return board_ids[0]

    def create_card_data_for_managed_copy(self, card: trello.Card):
        """Create a dictionary of card constructor arguments to make a card duplicate."""
        card_data = {}
        card_data['name'] = f'{_MANIPULATOR_SYMBOL} {self._boards_by_id[card.board_id].name}' \
            f' {_MANIPULATOR_SYMBOL} {card.name}'
        card_data['desc'] = f'{_MANIPULATOR_SYMBOL} this is auto-created card linked to' \
            f' {card.url} from board {self._boards_by_id[card.board_id].url}' \
            f' {_MANIPULATOR_SYMBOL}\n\n{card.description}'
        if card.due_date:
            try:
                card_data['due'] = card.due_date.isoformat()
            except AttributeError:
                _LOG.exception('failed to obtain due_date (type: %s, value: %s) for card %s',
                               type(card.due_date), card.due_date, card_data)
        # 'labels': []
        # 'position': 'bottom'
        return card_data

    def set_warmup_board(self, name: str) -> None:
        board_id = self.find_board_id('name', name)
        board = BoardCache(self.get_board(board_id), BoardKind.Warmup)
        self._warmup_board = board
        _LOG.info('set warmup board to %s', board)

    def set_work_board(self, name: str) -> None:
        board_id = self.find_board_id('name', name)
        board = BoardCache(self.get_board(board_id), BoardKind.Work)
        self._work_board = board
        _LOG.info('set Work board to %s', board)

    def set_handled_normal_boards(self, board_names: t.Iterable[str]) -> None:
        """Set boards that should be assumed to contain non-recurring tasks."""
        boards = []
        for name in board_names:
            board_id = self.find_board_id('name', name)
            board = BoardCache(self.get_board(board_id), BoardKind.Normal)
            boards.append(board)
            _LOG.info('set handled normal board %s', board)
        self._boards_normal = boards

    def set_handled_recurring_boards(self, board_names: t.Iterable[str]) -> None:
        """Set boards that should be assumed to contain recurring tasks."""
        boards = []
        for name in board_names:
            board_id = self.find_board_id('name', name)
            board = RecurringBoardCache(self.get_board(board_id), BoardKind.Recurring)
            boards.append(board)
            _LOG.info('set handled recurring board %s', board)
        self._boards_recurring = boards

    def set_boards_automatically(self) -> None:
        """Detect boards to be used by autotrello using configured name suffixes."""
        self.refresh_boards_list()
        normal_boards: t.List[str] = []
        recurring_boards: t.List[str] = []
        for board in self.boards:
            if board.name.endswith(BOARD_SYMBOLS[BoardKind.Warmup]):
                self.set_warmup_board(board.name)
            elif board.name.endswith(BOARD_SYMBOLS[BoardKind.Work]):
                self.set_work_board(board.name)
            elif board.name.endswith(BOARD_SYMBOLS[BoardKind.Normal]):
                normal_boards.append(board.name)
            elif board.name.endswith(BOARD_SYMBOLS[BoardKind.Recurring]):
                recurring_boards.append(board.name)
        self.set_handled_normal_boards(normal_boards)
        self.set_handled_recurring_boards(recurring_boards)

    def start_warmup(self) -> None:
        """Start warmup."""
        _LOG.info('starting warmup...')
        # create a list of card names being worked on
        self._work_board.refresh_cache()
        cards_being_worked_on = {card.name for card in self._work_board.cached_cards['Doing']}

        self._warmup_board.refresh_cache()
        for list_name in BOARD_LISTS[BoardKind.Warmup]:
            assert not self._warmup_board.cached_cards[list_name], list_name
        # for list_name, cards in self._warmup_board.cached_cards.items():
        #    if list_name == 'Backlog':
        #        _LOG.info('%i cards in backlog', len(cards))
        #        _LOG.debug('cards in backlog: %s', cards)
        #        continue
        #    if len(cards) > 0:
        #        _LOG.error('cannot start my day: cards in %s in list %s.',
        #                   self._startmyday_board, list_name)
        #        raise RuntimeError()

        for board in self._boards_recurring:
            _LOG.info('organizing recurring board %s', board)
            board.organize_cards()

        # add cards from normal boards and recurring boards
        for board in itertools.chain(self._boards_normal, self._boards_recurring):
            _LOG.info('scanning normal board %s', board)
            for card in board.cached_cards['To do']:
                card_data = self.create_card_data_for_managed_copy(card)
                if card_data['name'] in cards_being_worked_on:
                    list_name = 'Now'
                else:
                    list_name = '?'
                _LOG.info('creating card: "%s" in list "%s"', card_data['name'], list_name)
                _LOG.debug('created card data: %s', card_data)
                self._warmup_board.cached_lists[list_name].add_card(**card_data)
        _LOG.info('started warmup.')

    def abort_warmup(self) -> None:
        """Abort warm up session if it is ongoing, do nothing otherwise."""
        if self._warmup_board is None:
            return
        self._warmup_board.refresh_cache()
        for list_name in BOARD_LISTS[BoardKind.Warmup]:
            for card in self._warmup_board.cached_cards[list_name]:
                assert card.name.startswith(_MANIPULATOR_SYMBOL), card
                _LOG.warning('deleting card %s', card)
                card.delete()

    def start_work(self) -> None:
        """Start work, assuming that all cards in warmup are properly classified."""
        _LOG.info('starting work...')
        self._warmup_board.refresh_cache()
        # cannot start work if some tasks are not moved from ? into Now/Later/Never
        if self._warmup_board.cached_cards['?']:
            raise ValueError('List "?" on warmup board has {} cards left, please move them.'.format(
                len(self._warmup_board.cached_cards['?'])))
        self._work_board.refresh_cache()
        cards_being_worked_on = {card.name for card in self._work_board.cached_cards['Doing']}
        self.abort_work()
        # for list_name in BOARD_LISTS[BoardKind.Work]:
        #    if self._work_board.cached_cards[list_name]:
        #        raise ValueError(
        #            'List "{}" on work board has {} cards left, please abort work.'.format(
        #                list_name, len(self._warmup_board.cached_cards[list_name])))
        # delete auto-generated cards that are classified as Later
        for card in self._warmup_board.cached_cards['Later']:
            assert card.name.startswith(_MANIPULATOR_SYMBOL), card
            _LOG.warning('deleting card %s', card)
            card.delete()
        # delete auto-generated cards that are classified as Never
        # for card in self._warmup_board.cached_cards['Never']:
        #    assert card.name.startswith(_MANIPULATOR_SYMBOL), card
        #    _LOG.warning('deleting card %s', card)
        #    card.delete()
        # add Doing cards from normal boards
        for board in self._boards_normal:
            for card in board.cached_cards['Doing']:
                card_data = self.create_card_data_for_managed_copy(card)
                if card_data['name'] in cards_being_worked_on:
                    list_name = 'Doing'
                else:
                    list_name = 'To do'
                _LOG.info('creating card: "%s" in list "%s"', card_data['name'], list_name)
                _LOG.debug('created card data: %s', card_data)
                self._work_board.cached_lists[list_name].add_card(**card_data)
        # move cards classified as Now
        for card in self._warmup_board.cached_cards['Now']:
            assert card.name.startswith(_MANIPULATOR_SYMBOL), card
            _LOG.info('moving card %s to board %s', card, self._work_board)
            if card.name in cards_being_worked_on:
                list_name = 'Doing'
            else:
                list_name = 'To do'
            card.change_board(self._work_board.id, self._work_board.cached_lists[list_name].id)
        # _LOG.info('list %s has %i cards', list_name, len(cards))
        # _LOG.debug('list %s has cards: %s', list_name, cards)
        _LOG.info('work started.')

    def abort_work(self) -> None:
        """Abort today work if it is ongoing, do nothing otherwise."""
        if self._work_board is None:
            return
        self._work_board.refresh_cache()
        for list_name in BOARD_LISTS[BoardKind.Work]:
            cards = self._work_board.cached_cards[list_name]
            # _LOG.info('list %s has %i cards', list_name, len(cards))
            # _LOG.debug('list %s has cards: %s', list_name, cards)
            for card in cards:
                assert card.name.startswith(_MANIPULATOR_SYMBOL), card
                _LOG.warning('deleting card %s', card)
                card.delete()



"""Initialization of tests of autotrello package."""

import logging

import colorlog

_HANDLER = logging.StreamHandler()
_HANDLER.setFormatter(colorlog.ColoredFormatter(
    '{name} [{log_color}{levelname}{reset}] {message}', style='{'))

logging.basicConfig(level=logging.DEBUG, handlers=[_HANDLER])
logging.getLogger().setLevel(logging.WARNING)
logging.getLogger('autotrello').setLevel(logging.DEBUG)
logging.getLogger('test').setLevel(logging.DEBUG)



"""Initialization of autotrello package."""

__all__ = ['BoardCache', 'RecurringBoardCache', 'TrelloManipulator']

from .board_cache import BoardCache
from .recurring_board_cache import RecurringBoardCache
from .trello_manipulator import TrelloManipulator



"""Tests of autotrello.trello_manipulator."""

import os
import unittest

from autotrello.trello_manipulator import TrelloManipulator


@unittest.skipUnless(
    os.environ.get('AUTOTRELLO_TRELLO_API_KEY'), 'skipping tests depending on Trello API key')
class Tests(unittest.TestCase):

    def test_construct(self):
        trm = TrelloManipulator(api_key=os.environ['AUTOTRELLO_TRELLO_API_KEY'],
                                token=os.environ['AUTOTRELLO_TRELLO_TOKEN'])
        self.assertIsInstance(trm, TrelloManipulator)

    def test_refresh_boards_list(self):
        trm = TrelloManipulator(api_key=os.environ['AUTOTRELLO_TRELLO_API_KEY'],
                                token=os.environ['AUTOTRELLO_TRELLO_TOKEN'])
        trm.refresh_boards_list()
        self.assertGreater(len(trm.boards), 0)

    def test_set_warmup_board(self):
        trm = TrelloManipulator(api_key=os.environ['AUTOTRELLO_TRELLO_API_KEY'],
                                token=os.environ['AUTOTRELLO_TRELLO_TOKEN'])
        trm.refresh_boards_list()
        trm.set_warmup_board('Warm up ⭤')
        self.assertIsNotNone(trm.warmup_board)

    def test_set_work_board(self):
        trm = TrelloManipulator(api_key=os.environ['AUTOTRELLO_TRELLO_API_KEY'],
                                token=os.environ['AUTOTRELLO_TRELLO_TOKEN'])
        trm.refresh_boards_list()
        trm.set_work_board('Today 🗓')
        self.assertIsNotNone(trm.work_board)

    @unittest.skipUnless(os.environ.get('TEST_LONG'), 'skipping long test')
    def test_set_boards_automatically(self):
        trm = TrelloManipulator(api_key=os.environ['AUTOTRELLO_TRELLO_API_KEY'],
                                token=os.environ['AUTOTRELLO_TRELLO_TOKEN'])
        trm.set_boards_automatically()
        self.assertIsNotNone(trm.warmup_board)
        self.assertIsNotNone(trm.work_board)
        self.assertGreater(len(trm.normal_boards), 0)
        self.assertGreater(len(trm.recurring_boards), 0)

    def test_refresh_cache(self):
        trm = TrelloManipulator(api_key=os.environ['AUTOTRELLO_TRELLO_API_KEY'],
                                token=os.environ['AUTOTRELLO_TRELLO_TOKEN'])
        trm.refresh_boards_list()
        trm.set_warmup_board('Warm up ⭤')
        trm.set_work_board('Today 🗓')
        trm.set_handled_normal_boards(['autotrello ⭲'])
        self.assertGreater(len(trm.normal_boards), 0)
        self.assertEqual(trm.normal_boards[0].name, 'autotrello ⭲')
        trm.refresh_cache()
        self.assertGreater(len(trm.normal_boards[0].cached_lists), 0)
        self.assertGreater(len(trm.normal_boards[0].cached_cards), 0)

    @unittest.skipUnless(os.environ.get('TEST_LONG'), 'skipping long test')
    def test_full_sync(self):
        trm = TrelloManipulator(api_key=os.environ['AUTOTRELLO_TRELLO_API_KEY'],
                                token=os.environ['AUTOTRELLO_TRELLO_TOKEN'])
        trm.set_boards_automatically()
        self.assertGreater(len(trm.normal_boards), 0)
        self.assertGreater(len(trm.recurring_boards), 0)
        trm.refresh_cache()



"""Trello Board with cache."""

import collections
import enum
import logging
import typing as t

import trello  # py-trello package

_LOG = logging.getLogger(__name__)


@enum.unique
class BoardKind(enum.Enum):
    """Available kinds of boards."""

    Warmup = 1 + 4
    Work = 1 + 8
    Normal = 2 + 16
    Recurring = 2 + 32


BOARD_SYMBOLS = {
    BoardKind.Warmup: '⭤',
    BoardKind.Work: '🗓',
    BoardKind.Normal: '⭲',
    BoardKind.Recurring: '⭮'}


BOARD_LISTS = {
    BoardKind.Warmup: ['?', 'Now', 'Later'],
    BoardKind.Work: ['To do', 'Doing', 'Done', 'Not today'],
    BoardKind.Normal: ['To do', 'Doing'],
    BoardKind.Recurring: ['To do', 'Done']}


class BoardCache:
    """Wrapper around trello.Board that caches all downloaded data."""

    def __init__(self, board: trello.Board, board_kind: BoardKind):
        # super().__init__(*args, **kwargs)
        self._board = board
        self._board_kind = board_kind
        self._lists = {}  # type: t.Mapping[str, trello.List]
        self._list_cards = {}  # type: t.Mapping[str, t.Sequence[trello.Card]]
        self._refresh()

    @property
    def cached_lists(self) -> t.Mapping[str, trello.List]:
        return self._lists

    @property
    def cached_cards(self) -> t.Mapping[str, t.Sequence[trello.Card]]:
        return self._list_cards

    def cached_list(self, name: str) -> trello.List:
        return self._lists[name]

    def cached_list_cards(self, name: str) -> t.Sequence[trello.Card]:
        return self._list_cards[name]

    def refresh_cache(self):
        self._refresh()

    def _refresh(self):
        if not self._board.name.endswith(BOARD_SYMBOLS[self._board_kind]):
            raise ValueError(
                'Board "{}" is invalid: name of board of kind {} should end with {}.'
                .format(self._board.name, self._board_kind, BOARD_SYMBOLS[self._board_kind]))
        self._lists = collections.OrderedDict()
        self._list_cards = collections.OrderedDict()
        for board_list in self._board.all_lists():
            name = board_list.name
            if name not in BOARD_LISTS[self._board_kind]:
                continue
            self._lists[name] = board_list
            self._list_cards[name] = board_list.list_cards()
        for list_name in BOARD_LISTS[self._board_kind]:
            if list_name not in self._lists:
                raise ValueError('Board "{}" is invalid: board of kind {} should contain list "{}".'
                                 .format(self._board.name, self._board_kind, list_name))

    def __getattr__(self, name):
        return getattr(self._board, name)

    def __repr__(self):
        return repr(self._board)

    def __str__(self):
        return str(self._board)



"""Setup script for autotrello package."""

import setup_boilerplate


class Package(setup_boilerplate.Package):
    """Package metadata."""

    name = 'autotrello'
    description = 'Automation of workflows and high-level access to Trello.'
    url = 'https://github.com/mbdevpl/autotrello'
    classifiers = [
        'Development Status :: 1 - Planning',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Utilities']
    keywords = ['automation', 'trello']


if __name__ == '__main__':
    Package.setup()



"""Trello card linked to another Trello card."""

import trello  # py-trello package


class LinkedCard:  # pylint: disable = too-few-public-methods
    """Trello card linked to another for easy synchronization."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.card: trello.Card = None



"""Trello board for recurring tasks."""

import datetime
import enum
import itertools
import logging
import math
import typing as t

import trello  # py-trello package

from .board_cache import BoardCache

WEEK = 7.0
MONTH = 30.4375
YEAR = 365.25


@enum.unique
class Every(enum.Enum):
    """Available frequencies."""

    OneDay = 1
    TwoDays = 2
    ThreeDays = 3
    OneWeek = 7
    TwoWeeks = 14
    ThreeWeeks = 21
    OneMonth = 30
    TwoMonths = 60
    ThreeMonths = 90
    Quarter = 91
    HalfYear = 182
    OneYear = 365
    TwoYears = 731
    ThreeYears = 1096


FREQUENCIES = {
    Every.OneDay: datetime.timedelta(days=1),
    Every.TwoDays: datetime.timedelta(days=2),
    Every.ThreeDays: datetime.timedelta(days=3),
    Every.OneWeek: datetime.timedelta(days=round(1 * WEEK)),
    Every.TwoWeeks: datetime.timedelta(days=round(2 * WEEK)),
    Every.ThreeWeeks: datetime.timedelta(days=round(3 * WEEK)),
    Every.OneMonth: datetime.timedelta(days=round(1 * MONTH)),
    Every.TwoMonths: datetime.timedelta(days=round(2 * MONTH)),
    Every.ThreeMonths: datetime.timedelta(days=round(3 * MONTH)),
    Every.Quarter: datetime.timedelta(days=round(0.25 * YEAR)),
    Every.HalfYear: datetime.timedelta(days=round(0.5 * YEAR)),
    Every.OneYear: datetime.timedelta(days=round(1 * YEAR)),
    Every.TwoYears: datetime.timedelta(days=round(2 * YEAR)),
    Every.ThreeYears: datetime.timedelta(days=round(3 * YEAR))}

REMINDERS = {id_: datetime.timedelta(days=round(math.sqrt(freq.days)))
             for id_, freq in FREQUENCIES.items()}

FREQUENCIES_TRANSLATIONS = {
    'en': {
        'daily': Every.OneDay,
        'weekly': Every.OneWeek,
        'biweekly': Every.TwoWeeks,
        'triweekly': Every.ThreeWeeks,
        'monthly': Every.OneMonth,
        'bimonthly': Every.TwoMonths,
        'trimonthly': Every.ThreeMonths,
        'quarterly': Every.Quarter,
        'semianually': Every.HalfYear,
        'yearly': Every.OneYear,
        'biyearly': Every.TwoYears,
        'triyearly': Every.ThreeYears},
    'pl': {
        'codziennie': Every.OneDay,
        'co 2 dni': Every.TwoDays,
        'co 3 dni': Every.ThreeDays,
        'co tydzień': Every.OneWeek,
        'co 2 tyg.': Every.TwoWeeks,
        'co 3 tyg.': Every.ThreeWeeks,
        'co miesiąc': Every.OneMonth,
        'co 2 mies.': Every.TwoMonths,
        'co 3 mies.': Every.ThreeMonths,
        'co kwartał': Every.Quarter,
        'co pół roku': Every.HalfYear,
        'co rok': Every.OneYear,
        'co 2 lata': Every.TwoYears,
        'co 3 lata': Every.ThreeYears},
    'ja': {
        '毎日': Every.OneDay,
        '隔日': Every.TwoDays,
        '３日毎': Every.ThreeDays,
        '毎週': Every.OneWeek,
        '隔週': Every.TwoWeeks,
        '３週毎': Every.ThreeWeeks,
        '毎月': Every.OneMonth,
        '隔月': Every.TwoMonths,
        '３月毎': Every.ThreeMonths,
        '四半期毎': Every.Quarter,
        '半年毎': Every.HalfYear,
        '毎年': Every.OneYear,
        '隔年': Every.TwoYears,
        '３年毎': Every.ThreeYears}}

FREQUENCIES_MEANINGS = {}
for _, translations in FREQUENCIES_TRANSLATIONS.items():
    FREQUENCIES_MEANINGS.update(translations)

_LOG = logging.getLogger(__name__)


class RecurringBoardCache(BoardCache):
    """A cache for trello.Board capable of automatic card organization based on some assumptions.

    Assumptions:

    1. columns 'To do' and 'Done' exist

    2. all cards in those columns have due dates

    3. names of all cards in those columns start with one of predefined prefixes
    as defined in FREQUENCIES_TRANSLATIONS

    If all conditions are satisfied, auto-organization can happen.

    All cards with near due date are moved to To do, and all cards with not near due date are moved
    to Done.

    Nearness is determined based on frequency and deadline information,
    formula given by FREQUENCIES and REMINDERS variables and _refresh() method.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._now = None
        self._soon: t.List[trello.Card] = []
        self._not_soon: t.List[trello.Card] = []

    def _refresh(self):
        super()._refresh()
        self._now = datetime.datetime.now().astimezone()

        invalid = []
        self._soon = []
        self._not_soon = []
        for card in itertools.chain(self.cached_cards['To do'], self.cached_cards['Done']):
            if not card.due_date:
                _LOG.warning('skipping card %s: no due date!', card)
                invalid.append(card)
                continue
            if ':' in card.name:
                card_frequency_str = card.name.partition(':')[0]
                if card_frequency_str not in FREQUENCIES_MEANINGS:
                    _LOG.warning('skipping card %s: frequency "%s" unrecognized!',
                                 card, card_frequency_str)
                    invalid.append(card)
                    continue
            else:
                _LOG.warning('skipping card %s: no frequency information!', card)
                invalid.append(card)
                continue
            if card.due_date - REMINDERS[FREQUENCIES_MEANINGS[card_frequency_str]] <= self._now:
                self._soon.append(card)
            else:
                self._not_soon.append(card)
        if invalid:
            raise ValueError('Following cards are invalid for recurring board: {}'.format(
                [card.name for card in invalid]))
        self._soon = sorted(self._soon, key=lambda x: x.due_date)
        self._not_soon = sorted(self._not_soon, key=lambda x: x.due_date)

    def organize_cards(self):
        """Move cards between To do and Done columns depending on their deadlines."""
        todo_list = self.cached_lists['To do']
        done_list = self.cached_lists['Done']
        moved_to_todo = False
        for card in self._soon:
            if card.list_id == todo_list.id:
                continue
            _LOG.warning('moving card %s to %s', card, todo_list)
            card.change_list(todo_list.id)
            moved_to_todo = True
        moved_to_done = False
        for card in self._not_soon:
            if card.list_id == done_list.id:
                continue
            _LOG.warning('moving card %s to %s', card, done_list)
            card.change_list(done_list.id)
            moved_to_done = True
        if moved_to_todo:
            _LOG.warning('sorting cards in %s', todo_list)
            for card in reversed(self._soon):
                card.change_pos('top')
        if moved_to_done:
            _LOG.warning('sorting cards in %s', done_list)
            for card in reversed(self._not_soon):
                card.change_pos('top')
        if moved_to_todo or moved_to_done:
            self._refresh()
